import re
from logging import getLogger

from defence360agent.contracts.config import Malware as Config
from defence360agent.contracts.license import LicenseError
from defence360agent.contracts.messages import MalwareMRSUpload, MalwareScan
from defence360agent.contracts.plugins import (
    MessageSink,
    MessageSource,
    expect,
)
from defence360agent.malwarelib.utils import malware_response

logger = getLogger(__name__)


class SuspiciousToMRSUploader(MessageSink, MessageSource):
    ERR_MSG = "Failed to submit a suspicious"
    SUSP_PATTERN = re.compile(r"(?:suspicious\..+|[CS]MW-SUS-.+|SMW-HEUR-ELF)")

    async def create_source(self, loop, sink):
        self._sink = sink

    async def create_sink(self, loop):
        pass

    @classmethod
    def _check_hit(cls, hit):
        return hit["suspicious"] and cls.SUSP_PATTERN.match(hit["matches"])

    @expect(MalwareScan)
    async def process_scan(self, message):
        results = message["results"]
        if results is None:
            return
        if not Config.SEND_FILES:
            logger.info("Uploading of suspicious files to MRS is disabled")
            return
        suspicious_hits = [
            file
            for file, data in results.items()
            if any(self._check_hit(hit) for hit in data["hits"])
        ]
        if suspicious_hits:
            # Move uploading to another task
            await self._sink.process_message(
                MalwareMRSUpload({"suspicious_hits": suspicious_hits})
            )

    @expect(MalwareMRSUpload)
    async def process_hits(self, message):
        suspicious_hits = message["suspicious_hits"]
        upload_reason = message.get("upload_reason", "suspicious")
        for file in suspicious_hits:
            try:
                await malware_response.upload_with_retries(
                    file,
                    upload_reason=upload_reason,
                )
            except LicenseError as e:
                logger.warning("Cannot process message %s: %s", message, e)
                break
            except FileNotFoundError as e:
                err = "{}. {}".format(self.ERR_MSG, e.strerror)
                logger.warning("%s: %s", err, e.filename)
